@@ -39,6 +39,7 @@ class Scrambler(Module):
     cf section 5.2.3
     """
     def __init__(self, data_width, seed=0x7f80):
+        self.enable  = Signal(reset=1)
         self.sink    = sink   = Record([("data", data_width)])
         self.source  = source = Record([("data", data_width)])
         self.valid   = Signal()
@@ -60,8 +61,12 @@ class Scrambler(Module):
         source.data.reset_less = True
         self.sync += [
             self.valid.eq(1),
-            source.data.eq(swizzle(feedback, data_width)),
-            state.eq(full)
+            state.eq(full),
+            If(self.enable,
+                source.data.eq(swizzle(feedback, data_width))
+            ).Else(
+                source.data.eq(sink.data)
+            )
         ]
 
 
@@ -71,6 +76,7 @@ class Descrambler(Scrambler):
     cf section 5.2.3
     """
     def __init__(self, data_width, seed=0x7f80):
+        self.enable  = Signal(reset=1)
         self.sink    = sink   = Record([("data", data_width)])
         self.source  = source = Record([("data", data_width)])
         self.valid   = Signal()
@@ -92,8 +98,12 @@ class Descrambler(Scrambler):
         source.data.reset_less = True
         self.sync += [
             self.valid.eq(1),
-            source.data.eq(swizzle(feedback, data_width)),
-            state.eq(full)
+            state.eq(full),
+            If(self.enable,
+                source.data.eq(swizzle(feedback, data_width))
+            ).Else(
+                source.data.eq(sink.data)
+            )
         ]
 
 
@@ -101,9 +111,9 @@ class Descrambler(Scrambler):
 
 @ResetInserter()
 class Framer(Module):
-    """Framer
-    """
+    """Framer"""
     def __init__(self, data_width, octets_per_frame, frames_per_multiframe):
+        self.enable  = Signal(reset=1)
         self.sink    = sink   = Record([("data", data_width)])
         self.source  = source = Record(link_layout(data_width))
         self.latency = 0
@@ -114,11 +124,11 @@ class Framer(Module):
         frames_per_clock      = data_width//frame_width
         clocks_per_multiframe = frames_per_multiframe//frames_per_clock
 
-        # at least a frame per clock
+        # At least a frame per clock
         assert frame_width <= data_width
-        # multiple number of frame per clock
+        # Multiple number of frame per clock
         assert data_width%frame_width == 0
-        # multiframes aligned on clock
+        # Multiframes aligned on clock
         assert frames_per_multiframe%frames_per_clock == 0
 
         frame_last = 0
@@ -127,25 +137,22 @@ class Framer(Module):
                 frame_last |= (1<<i)
 
         counter = Signal(8)
-        self.sync += \
-            If(source.multiframe_last != 0,
-                counter.eq(0)
-            ).Else(
-                counter.eq(counter+1)
-            )
+        self.sync += counter.eq(counter+1)
+        self.sync += If(source.multiframe_last != 0, counter.eq(0))
 
         self.comb += [
             source.data.eq(sink.data),
-            source.frame_last.eq(frame_last),
-            If(counter == (clocks_per_multiframe-1),
-                source.multiframe_last.eq(1<<(data_width//8)-1)
+            If(self.enable,
+                source.frame_last.eq(frame_last),
+                If(counter == (clocks_per_multiframe-1),
+                    source.multiframe_last.eq(1<<(data_width//8)-1)
+                )
             )
         ]
 
 @ResetInserter()
 class Deframer(Module):
-    """Deframer
-    """
+    """Deframer"""
     def __init__(self, data_width, octets_per_frame, frames_per_multiframe):
         self.sink    = sink   = Record(link_layout(data_width))
         self.source  = source = Record([("data", data_width)])
@@ -165,9 +172,7 @@ class Deframer(Module):
         assert frames_per_multiframe%frames_per_clock == 0
 
         # FIXME: start on multiframe boundary?
-        self.comb += [
-            source.data.eq(sink.data),
-        ]
+        self.comb += source.data.eq(sink.data)
 
 # Alignment ----------------------------------------------------------------------------------------
 
@@ -176,29 +181,39 @@ class AlignInserter(Module):
     cf section 5.3.3.4.3
     """
     def __init__(self, data_width):
-        self.sink    = sink   = Record(link_layout(data_width))
-        self.source  = source = Record(link_layout(data_width))
+        self.scrambling = Signal(reset=1)
+        self.sink       = sink   = Record(link_layout(data_width))
+        self.source     = source = Record(link_layout(data_width))
         self.latency = 0
 
         # # #
 
         self.comb += source.eq(sink)
 
+        data_last = [Signal(8) for i in range(data_width//8)]
+
         for i in range(data_width//8):
+            data = sink.data[8*i:8*(i+1)]
             self.comb += [
-                # last scrambled octet in a multiframe equals "A" control character
-                If(sink.data[8*i:8*(i+1)] == control_characters["A"],
-                    If(sink.multiframe_last[i],
-                        source.ctrl[i].eq(1)
+                # With    Scrambling : If last octet in a multiframe equals "A" control character.
+                # Without Scrambling : If last octet in a multiframe equals previous octet.
+                If(sink.multiframe_last[i],
+                    If(( self.scrambling & (data == control_characters["A"])) |
+                       (~self.scrambling & (data == data_last[i])),
+                        source.ctrl[i].eq(1),
+                        source.data[8*i:8*(i+1)].eq(control_characters["A"]),
                     )
-                # last scrambled octet in a frame but not at the end of a
-                # multiframe equals "F" control character
-                ).Elif(sink.data[8*i:8*(i+1)] == control_characters["F"],
-                    If(sink.frame_last[i] & ~sink.multiframe_last[i],
-                        source.ctrl[i].eq(1)
+                # With    Scrambling : If last octet in a frame (not at a end of multiframe) equals "F" control character.
+                # Without Scrambling : If last octet in a frame (not at a end of multiframe) equals previous octet.
+                ).Elif(sink.frame_last[i],
+                    If(( self.scrambling & (data == control_characters["F"])) |
+                       (~self.scrambling & (data == data_last[i])),
+                        source.ctrl[i].eq(1),
+                        source.data[8*i:8*(i+1)].eq(control_characters["F"]),
                     )
                 )
             ]
+            self.sync += data_last[i].eq(data)
 
 
 class AlignReplacer(Module):
@@ -206,17 +221,37 @@ class AlignReplacer(Module):
     cf section 5.3.3.4.3
     """
     def __init__(self, data_width):
-        self.sink    = sink   = Record(link_layout(data_width))
-        self.source  = source = Record(link_layout(data_width))
+        self.scrambling = Signal(reset=1)
+        self.sink       = sink   = Record(link_layout(data_width))
+        self.source     = source = Record(link_layout(data_width))
         self.latency = 0
 
         # # #
 
-        # recopy scrambler data and set ctrl to 0
         self.comb += source.eq(sink)
-        for i in range(data_width//8):
-            self.comb += source.ctrl.eq(0)
 
+        data_last = [Signal(8) for i in range(data_width//8)]
+
+        for i in range(data_width//8):
+            data = sink.data[8*i:8*(i+1)]
+            self.comb += [
+                If(self.scrambling,
+                    # No Replacement, just set ctrl to 0.
+                    source.ctrl[i].eq(0)
+                ).Else(
+                    # If "A" control character, replace with last octet.
+                    If(sink.ctrl[i] & (sink.data[8*i:8*(i+1)] == control_characters["A"]),
+                        source.ctrl[i].eq(0),
+                        source.data[8*i:8*(i+1)].eq(data_last[i])
+                    ),
+                    # If "A" control character, replace with last octet.
+                    If(sink.ctrl[i] & (sink.data[8*i:8*(i+1)] == control_characters["F"]),
+                        source.ctrl[i].eq(0),
+                        source.data[8*i:8*(i+1)].eq(data_last[i])
+                    )
+                )
+            ]
+            self.sync += data_last[i].eq(data)
 
 class Aligner(Module):
     def __init__(self, data_width):
@@ -260,8 +295,7 @@ class Aligner(Module):
 # Code Group Synchronization -----------------------------------------------------------------------
 
 class CGSGenerator(Module):
-    """Code Group Synchronization
-    """
+    """Code Group Synchronization"""
     def __init__(self, data_width):
         self.source = source = Record(link_layout(data_width))
 
@@ -276,8 +310,7 @@ class CGSGenerator(Module):
             ]
 
 class CGSChecker(Module):
-    """Code Group Synchronization
-    """
+    """Code Group Synchronization"""
     def __init__(self, data_width):
         self.sink  = sink  = Record(link_layout(data_width))
         self.valid = valid = Signal()
@@ -303,13 +336,9 @@ class ILAS:
     """Initial Lane Alignment Sequence
     cf section 5.3.3.5
     """
-    def __init__(self, data_width,
-                 octets_per_frame,
-                 frames_per_multiframe,
-                 configuration_data,
-                 with_counter=True):
+    def __init__(self, data_width, octets_per_frame, frames_per_multiframe, configuration_data, with_counter=True):
 
-        # compute ILAS's octets
+        # Compute ILAS's octets
 
         octets_per_multiframe = octets_per_frame*frames_per_multiframe
 
@@ -327,7 +356,7 @@ class ILAS:
                 multiframe[2:2+len(configuration_data)] = configuration_data
             octets += multiframe
 
-        # pack ILAS's octets in a lookup table
+        # Pack ILAS's octets in a lookup table
 
         octets_per_clock = data_width//8
 
@@ -346,32 +375,25 @@ class ILAS:
             data_words.append(data_word)
             ctrl_words.append(ctrl_word)
 
-        assert len(data_words) == (octets_per_frame*
-                                        frames_per_multiframe*
-                                        4//octets_per_clock)
+        #print("ILAs datas/ctrls:")
+        #for data, ctrl in zip(data_words, ctrl_words):
+        #    print(f"data: 0x{data:08x}, ctrl: 0x{ctrl:02x}")
+
+        assert len(data_words) == (octets_per_frame*frames_per_multiframe*4//octets_per_clock)
 
 @ResetInserter()
 class ILASGenerator(ILAS, Module):
     """Initial Lane Alignment Sequence Generator
     cf section 5.3.3.5
     """
-    def __init__(self, data_width,
-                 octets_per_frame,
-                 frames_per_multiframe,
-                 configuration_data,
-                 with_counter=True):
+    def __init__(self, data_width, octets_per_frame, frames_per_multiframe, configuration_data, with_counter=True):
         self.source = source = Record(link_layout(data_width))
         self.done = Signal()
 
         # # #
 
-        # compute ILAS's data/ctrl words
-        ILAS.__init__(self,
-            data_width,
-            octets_per_frame,
-            frames_per_multiframe,
-            configuration_data,
-            with_counter)
+        # Compute ILAS's data/ctrl words
+        ILAS.__init__(self, data_width, octets_per_frame, frames_per_multiframe, configuration_data, with_counter)
 
         data_lut = Memory(data_width, len(self.data_words), init=self.data_words)
         data_port = data_lut.get_port(async_read=True)
@@ -381,7 +403,7 @@ class ILASGenerator(ILAS, Module):
         ctrl_port = ctrl_lut.get_port(async_read=True)
         self.specials += ctrl_lut, ctrl_port
 
-        # stream data/ctrl from lookup tables
+        # Stream data/ctrl from lookup tables
         counter = Signal(max=len(self.data_words)+1)
         self.comb += [
             source.last.eq(counter == (len(self.data_words)-1)),
@@ -390,18 +412,18 @@ class ILASGenerator(ILAS, Module):
             source.data.eq(data_port.dat_r),
             source.ctrl.eq(ctrl_port.dat_r)
         ]
-        self.sync += \
+        self.sync += [
             If(counter != len(self.data_words),
                 counter.eq(counter + 1)
             )
+        ]
 
-        # done
+        # Done
         self.comb += self.done.eq(counter == len(self.data_words))
 
 
 class ILASStartChecker(Module):
-    """Code Group Synchronization
-    """
+    """Code Group Synchronization"""
     def __init__(self, data_width):
         self.sink  = sink  = Record(link_layout(data_width))
         self.valid = valid = Signal()
@@ -424,29 +446,20 @@ class ILASChecker(ILAS, Module):
     """Initial Lane Alignment Sequence Checker
     cf section 5.3.3.5
     """
-    def __init__(self, data_width,
-                 octets_per_frame,
-                 frames_per_multiframe,
-                 configuration_data,
-                 with_counter=True):
+    def __init__(self, data_width, octets_per_frame, frames_per_multiframe, configuration_data, with_counter=True):
         self.sink  = sink  = Record(link_layout(data_width))
         self.done  = done  = Signal()
         self.valid = valid = Signal()
 
         # # #
 
-        # detect start
+        # Detect start
         start =  ILASStartChecker(data_width)
         self.submodules.start = start
         self.comb += start.sink.eq(sink)
 
-        # compute ILAS's data/ctrl words
-        ILAS.__init__(self,
-            data_width,
-            octets_per_frame,
-            frames_per_multiframe,
-            configuration_data,
-            with_counter)
+        # Compute ILAS's data/ctrl words
+        ILAS.__init__(self, data_width, octets_per_frame, frames_per_multiframe, configuration_data, with_counter)
 
         data_lut  = Memory(data_width, len(self.data_words), init=self.data_words)
         data_port = data_lut.get_port(async_read=True)
@@ -459,7 +472,7 @@ class ILASChecker(ILAS, Module):
         self.data = data_port.dat_r
         self.ctrl = ctrl_port.dat_r
 
-        # compare data/ctrl with lookup tables
+        # Compare data/ctrl with lookup tables
         counter = Signal(max=len(self.data_words)+1)
         self.comb += [
             data_port.adr.eq(counter),
@@ -480,7 +493,7 @@ class ILASChecker(ILAS, Module):
             )
         ]
 
-        # done
+        # Done
         self.comb += self.done.eq(counter == len(self.data_words))
 
 # Link TX ------------------------------------------------------------------------------------------
@@ -493,18 +506,15 @@ class LiteJESD204BLinkTXDapath(Module):
         # # #
 
         # Scrambling
-        scrambler = Scrambler(data_width)
-        self.submodules.scrambler = scrambler
+        self.submodules.scrambler = scrambler = Scrambler(data_width)
 
         # Framing
-        framer = Framer(data_width,
-            octets_per_frame,
-            frames_per_multiframe)
+        framer = Framer(data_width, octets_per_frame, frames_per_multiframe)
         self.submodules.framer = framer
 
         # Alignment
-        align_inserter = AlignInserter(data_width)
-        self.submodules.align_inserter = align_inserter
+        self.submodules.align_inserter = align_inserter = AlignInserter(data_width)
+        self.comb += align_inserter.scrambling.eq(scrambler.enable)
 
         # Flow
         self.latency = scrambler.latency + framer.latency + align_inserter.latency
@@ -519,13 +529,12 @@ class LiteJESD204BLinkTXDapath(Module):
 
 @ResetInserter()
 class LiteJESD204BLinkTX(Module):
-    """Link TX layer
-    """
+    """Link TX layer"""
     def __init__(self, data_width, jesd_settings, n=0):
-        self.jsync     = Signal() # input
-        self.jref      = Signal() # input
-        self.lmfc_zero = Signal() # input
-        self.ready     = Signal() # output
+        self.jsync     = Signal() # Input
+        self.jref      = Signal() # Input
+        self.lmfc_zero = Signal() # Input
+        self.ready     = Signal() # Output
 
         self.sink   = sink   = Record([("data", data_width)])
         self.source = source = Record(link_layout(data_width))
@@ -549,9 +558,10 @@ class LiteJESD204BLinkTX(Module):
             jesd_settings.transport.k)
         self.submodules.datapath = datapath
         self.comb += datapath.sink.eq(sink)
+        self.comb += datapath.framer.enable.eq(int(jesd_settings.framing))
 
         # Sync
-        jsync_timer = WaitTimer(4) # distinguish errors reporting / re-synchronization requests.
+        jsync_timer = WaitTimer(4) # Distinguish errors reporting / re-synchronization requests.
         self.submodules += jsync_timer
         self.comb += jsync_timer.wait.eq(~self.jsync)
 
@@ -581,7 +591,7 @@ class LiteJESD204BLinkTX(Module):
             source.eq(datapath.source),
             If(jsync_timer.done,
                 NextState("SEND-CGS")
-            ),
+            )
         )
 
 # Link RX ------------------------------------------------------------------------------------------
@@ -594,8 +604,7 @@ class LiteJESD204BLinkRXDapath(Module):
         # # #
 
         # Alignment
-        align_replacer = AlignReplacer(data_width)
-        self.submodules.align_replacer = align_replacer
+        self.submodules.align_replacer = align_replacer = AlignReplacer(data_width)
 
         # Deframing
         deframer = Deframer(data_width,
@@ -604,8 +613,8 @@ class LiteJESD204BLinkRXDapath(Module):
         self.submodules.deframer = deframer
 
         # Descrambling
-        descrambler = Descrambler(data_width)
-        self.submodules.descrambler = descrambler
+        self.submodules.descrambler = descrambler = Descrambler(data_width)
+        self.comb += align_replacer.scrambling.eq(descrambler.enable)
 
         # Flow
         self.latency = align_replacer.latency + deframer.latency + descrambler.latency
@@ -619,14 +628,13 @@ class LiteJESD204BLinkRXDapath(Module):
 
 @ResetInserter()
 class LiteJESD204BLinkRX(Module):
-    """Link RX layer
-    """
+    """Link RX layer"""
     def __init__(self, data_width, jesd_settings, n=0, ilas_check=True):
-        self.jsync     = Signal() # output
-        self.jref      = Signal() # input
-        self.lmfc_zero = Signal() # input
-        self.ready     = Signal() # output
-        self.align     = Signal() # output
+        self.jsync     = Signal() # Output
+        self.jref      = Signal() # Input
+        self.lmfc_zero = Signal() # Input
+        self.ready     = Signal() # Output
+        self.align     = Signal() # Output
 
         self.sink   = sink   = Record(link_layout(data_width))
         self.source = source = Record([("data", data_width)])
